@@ -1,9 +1,11 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2010 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See COPYING file for details.
+ * modify it under the terms of the GNU General Public License version 2 
+ * or later as published by the Free Software Foundation. 
+ * See COPYING file for details.
  */
 
 /**
@@ -24,27 +26,55 @@
 struct Iso_Boot
 {
     IsoNode node;
+
+    /* Want to get content of loaded boot catalog.
+       Vreixo took care not to make it an IsoFile at load time.
+       So this is implemented independently of IsoStream.
+     */
+    uint32_t lba;
+    off_t size;
+    char *content;
 };
 
+/* Not more than 32 so that all entries fit into 2048 bytes */
+#define Libisofs_max_boot_imageS 32
+
 struct el_torito_boot_catalog {
     IsoBoot *node; /* node of the catalog */
-    struct el_torito_boot_image *image; /* default boot image */
+
+    int num_bootimages;
+    struct el_torito_boot_image *bootimages[Libisofs_max_boot_imageS];
+                                                  /* [0]= default boot image */
+
+    /* Weight value for image sorting */
+    int sort_weight;
 };
 
 struct el_torito_boot_image {
     IsoFile *image;
 
     unsigned int bootable:1; /**< If the entry is bootable. */
+    /**
+     * Whether the boot image seems to contain a boot_info_table
+     */
+    unsigned int seems_boot_info_table:1;
     /**
      * isolinux options
      * bit 0 -> whether to patch image
-     * bit 1 -> whether to create isolinux image
+     * bit 1 -> whether to put built-in isolinux 3.72 isohybrid-MBR into image
+     *          System Area (deprecated)
      */
     unsigned int isolinux_options:2;
     unsigned char type; /**< The type of image */
     unsigned char partition_type; /**< type of partition for HD-emul images */
     short load_seg; /**< Load segment for the initial boot image. */
     short load_size; /**< Number of sectors to load. */
+
+    /* Byte 1 of Validation Entry or Section Header Entry:
+       0= 80x86, 1= PowerPC, 2= Mac, 0xef= EFI */
+    uint8_t platform_id;
+    uint8_t id_string[28];
+    uint8_t selection_crit[20];
 };
 
 /** El-Torito, 2.1 */
@@ -74,8 +104,8 @@ struct el_torito_default_entry {
 struct el_torito_section_header {
     uint8_t header_indicator    BP(1, 1);
     uint8_t platform_id         BP(2, 2);
-    uint8_t number              BP(3, 4);
-    uint8_t character           BP(5, 32);
+    uint8_t num_entries         BP(3, 4);
+    uint8_t id_string           BP(5, 32);
 };
 
 /** El-Torito, 2.4 */
@@ -105,4 +135,14 @@ int el_torito_catalog_file_src_create(Ecma119Image *target, IsoFileSrc **src);
  */
 int eltorito_writer_create(Ecma119Image *target);
 
+/**
+ * Insert boot info table content into buf.
+ *
+ * @return
+ *      1 on success, 0 error (but continue), < 0 error
+ */
+int make_boot_info_table(uint8_t *buf, uint32_t pvd_lba,
+                         uint32_t boot_lba, uint32_t imgsize);
+
+
 #endif /* LIBISO_ELTORITO_H */
